{!!}

{
  Indy HL7 Minimal Lower Layer Protocol TIdHL7

    Original author Grahame Grieve

    This code was donated by HL7Connect.com
    For more HL7 open source code see
    http://www.hl7connect.com/tools

  This unit implements support for the Standard HL7 minimal Lower Layer
  protocol. For further details, consult the HL7 standard (www.hl7.org).

  Before you can use this component, you must set the following properties:
    CommunicationMode
    Address (if you want to be a client)
    Port
    isListener
  and hook the appropriate events (see below)

  This component will operate as either a server or a client depending on
  the configuration
}

{
 Version History:
   07/12/2001   Grahame Grieve      Various fixes for cmSingleThread mode
   05/11/2001   Grahame Grieve      Merge into Indy
   03/09/2001   Grahame Grieve      Prepare for Indy
}

(* note: Events are structurally important for this component. However there is
  a bug in SyncObjs for Linux under Kylix 1 and 2 where TEvent.WaitFor cannot be
  used with timeouts. If you compile your own RTL, then you can fix the routine
  like this:

    function TEvent.WaitFor(Timeout: LongWord): TWaitResult;
    {$IFDEF LINUX}
    var ts : TTimeSpec;
    begin
      ts.tv_sec  := timeout div 1000;
      ts.tv_nsec := (timeout mod 1000) * 1000000;
      if sem_timedwait(FSem, ts) = 0 then
        result := wrSignaled
      else
        result := wrTimeOut;
    {$ENDIF}

  and then disable this define: *)

{$IFDEF LINUX}
  {$DEFINE SYNCOBJS_WORKAROUND}
{$ENDIF}

  { this is a serious issue - unless you fix the RTL, this component does not
    function properly on Linux at the present time. This may be fixed in a
    future version }

unit IdHL7;

interface

uses
  Classes,
  IdBaseComponent,
  IdException,
  IdGlobal,
  IdTCPClient,
  IdTCPConnection,
  IdTCPServer,
  SyncObjs,
  SysUtils;

const
   MSG_START = #$0B;       {do not localize}
   MSG_END   = #$1C#$0D;   {do not localize}

   BUFFER_SIZE_LIMIT = 1024 * 1024;  // buffer is allowed to grow to this size without any
                                     // valid messages. Will be truncated with no notice (DoS protection)

type
   EHL7CommunicationError = class (EIdException)
    protected
      FInterfaceName: string;
    public
      constructor Create(AnInterfaceName, AMessage: string);
      property InterfaceName: string read FInterfaceName;
   end;


   THL7CommunicationMode = (
                    cmUnknown,        // not valid - default setting must be changed by application
                    cmAsynchronous,   // see comments below for meanings of the other parameters
                    cmSynchronous,
                    cmSingleThread);

   TSendResponse = (srNone,          // internal use only - never returned
                    srError,         // internal use only - never returned
                    srNoConnection,  // you tried to send but there was no connection
                    srSent,          // you asked to send without waiting, and it has been done
                    srOK,            // sent ok, and response returned
                    srTimeout);      // we sent but there was no response (connection will be dropped internally

   TIdHL7Status = (
                    isStopped,       // not doing anything
                    isNotConnected,  // not Connected (Server state)
                    isConnecting,    // Client is attempting to connect
                    isWaitReconnect, // Client is in delay loop prior to attempting to connect
                    isConnected,     // connected OK
                    isUnusable       // Not Usable - stop failed
                      );

const
   { default property values }
   DEFAULT_ADDRESS           = '';         {do not localize}
   DEFAULT_PORT              = 0;
   DEFAULT_TIMEOUT           = 30000;
   DEFAULT_RECEIVE_TIMEOUT   = 30000;
   NULL_IP                   = '0.0.0.0';  {do not localize}
   DEFAULT_CONN_LIMIT        = 1;
   DEFAULT_RECONNECT_DELAY   = 15000;
   DEFAULT_COMM_MODE         = cmUnknown;
   DEFAULT_IS_LISTENER       = true;

type
   TMessageArriveEvent  = procedure (Sender: TObject; Msg: string) of object;
   TMessageReceiveEvent = procedure (Sender: TObject; Msg: string; var Reply: string) of object;
   TReceiveErrorEvent   = procedure (Sender: TObject; Msg: string; e: Exception; var Reply: string; var DropConnection: boolean) of object;

   TIdHL7 = class;

   TIdHL7PeerThread = class (TIdPeerThread)
    protected
      FBuffer: string;
    public
      constructor Create(ACreateSuspended: Boolean = True); override;
      Destructor Destroy; override;
    end;

   TIdHL7ClientThread = class (TThread)
    protected
      FClient     : TIdTCPClient;
      FCloseEvent : TIdLocalEvent;
      FOwner      : TIdHL7;
      procedure Execute; override;
      procedure PollStack;
    public
      constructor Create(aOwner: TIdHL7);
      destructor Destroy; override;
    end;

   TIdHL7 = class (TIdBaseComponent)
      protected
        FLock               : TCriticalSection;
        FStatus             : TIdHL7Status;
        FStatusDesc         : string;

        // these queues hold messages when running in singlethread mode
        FMsgQueue            : TList;
        FHndMsgQueue         : TList;

        FAddress            : string;
        FCommunicationMode  : THL7CommunicationMode;
        FConnectionLimit    : integer;
        FIPMask             : string;
        FIPRestriction      : string;
        FIsListener         : boolean;
        FObject             : TObject;
        FPreStopped         : boolean;
        FPort               : word;
        FReconnectDelay     : cardinal;
        FTimeOut            : cardinal;
        FReceiveTimeout     : cardinal;


        FOnConnect          : TNotifyEvent;
        FOnDisconnect       : TNotifyEvent;
        FOnMessageArrive    : TMessageArriveEvent;
        FOnReceiveMessage   : TMessageReceiveEvent;
        FOnReceiveError     : TReceiveErrorEvent;

        FIsServer            : boolean;
        // current connection count (server only) (can only exceed 1 when mode is not
        // asynchronous and we are listening)
        FConnCount           : integer;
        FServer              : TIdTCPServer;
        // if we are a server, and the mode is not asynchronous, and we are not listening, then
        // we will track the current server connection with this, so we can initiate sending on it
        FServerConn          : TIdTCPServerConnection;

        // A thread exists to connect and receive incoming tcp traffic
        FClientThread        : TIdHL7ClientThread;
        FClient              : TIdTCPClient;

        // these fields are used for handling message response in synchronous mode
        FWaitingForAnswer    : boolean;
        FMsgReply            : string;
        FReplyResponse       : TSendResponse;
        FWaitEvent           : TIdLocalEvent;

        procedure SetAddress(const Value: string);
        procedure SetConnectionLimit(const Value: integer);
        procedure SetIPMask(const Value: string);
        procedure SetIPRestriction(const Value: string);
        procedure SetPort(const Value: word);
        procedure SetReconnectDelay(const Value: cardinal);
        procedure SetTimeOut(const Value: cardinal);
        procedure SetCommunicationMode(const Value: THL7CommunicationMode);
        procedure SetIsListener(const Value: boolean);
        function  GetStatus: TIdHL7Status;
        function  GetStatusDesc: string;

        procedure InternalSetStatus(const AStatus: TIdHL7Status; ADesc: string);

        procedure CheckServerParameters;
        procedure StartServer;
        procedure StopServer;
        procedure DropServerConnection;
        procedure ServerConnect(AThread: TIdPeerThread);
        procedure ServerExecute(AThread: TIdPeerThread);
        procedure ServerDisconnect(AThread: TIdPeerThread);

        procedure CheckClientParameters;
        procedure StartClient;
        procedure StopClient;
        procedure DropClientConnection;

        procedure HandleIncoming(var VBuffer: string; AConnection: TIdTCPConnection);
        function  HandleMessage(const AMsg: string; var VReply: string):boolean;
      public
        constructor Create(Component: TComponent); override;
        destructor Destroy; override;

        function  Going: boolean;

        // for the app to use to hold any related object
        property  ObjTag: TObject read FObject write FObject;

        // status
        property  Status: TIdHL7Status read GetStatus;
        property  StatusDesc: string read GetStatusDesc;
        function  Connected: boolean;

        property  IsServer: boolean read FIsServer;
        procedure Start;
        procedure PreStop; // call this in advance to start the shut down process. You do not need to call this
        procedure Stop;

        // asynchronous.
        function  AsynchronousSend(AMsg: string): TSendResponse;
        property  OnMessageArrive: TMessageArriveEvent read FOnMessageArrive write FOnMessageArrive;

        // synchronous
        function  SynchronousSend(AMsg: string; var VReply: string): TSendResponse;
        property  OnReceiveMessage: TMessageReceiveEvent read FOnReceiveMessage write FOnReceiveMessage;

        // single thread
        procedure SendMessage(AMsg: string);
        // you can't call SendMessage again without calling GetReply first
        function  GetReply(var VReply: string): TSendResponse;
        function  GetMessage(var VMsg: string): pointer;  // return nil if no messages
        // if you don't call SendReply then no reply will be sent.
        procedure SendReply(AMsgHnd: pointer; AReply: string);

      published
        // basic properties
        property  Address: string read FAddress write SetAddress;  // leave blank and we will be server
        property  Port: word read FPort write SetPort default DEFAULT_PORT;

        // milliseconds - message timeout - how long we wait for other system to reply
        property  TimeOut: cardinal read FTimeOut write SetTimeOut default DEFAULT_TIMEOUT;

        // milliseconds - message timeout. When running cmSingleThread, how long we wait for the application to process an incoming message before giving up
        property  ReceiveTimeout: cardinal read FReceiveTimeout write FReceiveTimeout default DEFAULT_RECEIVE_TIMEOUT;

        // server properties
        property  ConnectionLimit: integer read FConnectionLimit write SetConnectionLimit default DEFAULT_CONN_LIMIT; // ignored if isListener is false
        property  IPRestriction  : string read FIPRestriction write SetIPRestriction;
        property  IPMask         : string read FIPMask write SetIPMask;

        // client properties

        // milliseconds - how long we wait after losing connection to retry
        property  ReconnectDelay : cardinal read FReconnectDelay write SetReconnectDelay default DEFAULT_RECONNECT_DELAY;

        // message flow

        // Set this to one of 4 possibilities:
        //
        //    cmUnknown
        //       Default at start up. You must set a value before starting
        //
        //    cmAsynchronous
        //        Send Messages with AsynchronousSend. does not wait for
        //                   remote side to respond before returning
        //        Receive Messages with OnMessageArrive. Message may
        //                   be response or new message
        //       The application is responsible for responding to the remote
        //       application and dropping the link as required
        //       You must hook the OnMessageArrive Event before setting this mode
        //       The property IsListener has no meaning in this mode
        //
        //   cmSynchronous
        //       Send Messages with SynchronousSend. Remote applications response
        //                   will be returned (or timeout). Only use if IsListener is false
        //       Receive Messages with OnReceiveMessage. Only if IsListener is
        //                   true
        //       In this mode, the object will wait for a response when sending,
        //       and expects the application to reply when a message arrives.
        //       In this mode, the interface can either be the listener or the
        //       initiator but not both. IsListener controls which one.
        //       note that OnReceiveMessage must be thread safe if you allow
        //       more than one connection to a server
        //
        //   cmSingleThread
        //       Send Messages with SendMessage. Poll for answer using GetReply.
        //                   Only if isListener is false
        //       Receive Messages using GetMessage. Return a response using
        //                   SendReply. Only if IsListener is true
        //       This mode is the same as cmSynchronous, but the application is
        //       assumed to be single threaded. The application must poll to
        //       find out what is happening rather than being informed using
        //       an event in a different thread

        property  CommunicationMode: THL7CommunicationMode read FCommunicationMode write SetCommunicationMode default DEFAULT_COMM_MODE;

        // note that IsListener is not related to which end is client. Either end
        // may make the connection, and thereafter only one end will be the initiator
        // and one end will be the listener. Generally it is recommended that the
        // listener be the server. If the client is listening, network conditions
        // may lead to a state where the client has a phantom connection and it will
        // never find out since it doesn't initiate traffic. In this case, restart
        // the interface if there isn't traffic for a period
        property  IsListener: boolean read FIsListener write SetIsListener default DEFAULT_IS_LISTENER;

        // useful for application
        property OnConnect: TNotifyEvent read FOnConnect write FOnConnect;
        property OnDisconnect: TNotifyEvent read FOnDisconnect write FOnDisconnect;

        // this is called when an unhandled exception is generated by the
        // hl7 object or the application. It allows the application to
        // construct a useful return error, log the exception, and drop the
        // connection if it wants
        property OnReceiveError: TReceiveErrorEvent read FOnReceiveError write FOnReceiveError;
    end;

implementation

uses
  IdResourceStrings;

type
    TQueuedMessage = class
      FEvent: TIdLocalEvent;
      FMsg: string;
      FTimeOut: cardinal;
      FReply: string;
      procedure Wait;
      constructor Create(aMsg: string; ATimeOut : cardinal);
      destructor Destroy; override;
    end;

{ TQueuedMessage }

constructor TQueuedMessage.Create(aMsg: string; ATimeOut: cardinal);
begin
  inherited Create;
  FEvent := TIdLocalEvent.Create(false, false);
  FMsg := aMsg;
  FTimeOut := ATimeOut;
end;

destructor TQueuedMessage.Destroy;
begin
  FreeAndNil(FEvent);
  inherited;
end;

procedure TQueuedMessage.Wait;
begin
  {$IFDEF SYNCOBJS_WORKAROUND}
  FEvent.WaitFor;
  {$ELSE}
  FEvent.WaitFor(FTimeOut);
  {$ENDIF}
end;

{ EHL7CommunicationError }

constructor EHL7CommunicationError.Create(AnInterfaceName, AMessage: string);
begin
  FInterfaceName := AnInterfaceName;
  if FInterfaceName <> '' then         {do not localize}
  begin
    inherited Create('[' + AnInterfaceName + '] '+ AMessage)
  end
  else
  begin
    inherited Create(AMessage);
  end
end;

{ TIdHL7 }

constructor TIdHL7.Create;
begin
  inherited Create(Component);

  // partly redundant initialization of properties

  FIsListener        := DEFAULT_IS_LISTENER;
  FCommunicationMode := DEFAULT_COMM_MODE;
  FTimeOut           := DEFAULT_TIMEOUT;
  FReconnectDelay    := DEFAULT_RECONNECT_DELAY;
  FReceiveTimeout    := DEFAULT_RECEIVE_TIMEOUT;
  FConnectionLimit   := DEFAULT_CONN_LIMIT;
  FIPMask            := NULL_IP;
  FIPRestriction     := NULL_IP;
  FAddress           := DEFAULT_ADDRESS;
  FPort              := DEFAULT_PORT;
  FOnReceiveMessage  := nil;
  FOnConnect         := nil;
  FOnDisconnect      := nil;
  FObject            := nil;

  // initialise status
  FStatus            := IsStopped;
  FStatusDesc        := RSHL7StatusStopped;

  // build internal infrastructure
  Flock              := TCriticalSection.Create;
  FConnCount         := 0;
  FServer            := nil;
  FServerConn        := nil;
  FClientThread      := nil;
  FClient            := nil;
  FMsgQueue          := TList.Create;
  FHndMsgQueue       := TList.Create;
  FWaitingForAnswer  := false;
  FMsgReply          := '';   {do not localize}
  FReplyResponse     := srNone;
  FWaitEvent         := TIdLocalEvent.Create(false, false);
end;

destructor TIdHL7.Destroy;
begin
  try
    if Going then
    begin
      Stop;
    end;
  finally
    FreeAndNil(FMsgQueue);
    FreeAndNil(FHndMsgQueue);
    FreeAndNil(FWaitEvent);
    FreeAndNil(FLock);
    inherited;
  end;
end;

{==========================================================
  Property Servers
 ==========================================================}

procedure TIdHL7.SetAddress(const Value: string);
begin
  if Going then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWhileWorking, ['Address']));   {do not localize??}
  end;
  FAddress := Value;
end;

procedure TIdHL7.SetConnectionLimit(const Value: integer);
begin
  if Going then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWhileWorking, ['ConnectionLimit'])); {do not localize??}
  end;
  FConnectionLimit := Value;
end;

procedure TIdHL7.SetIPMask(const Value: string);
begin
  if Going then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWhileWorking, ['IP Mask']));  {do not localize??}
  end;
  FIPMask := Value;
end;

procedure TIdHL7.SetIPRestriction(const Value: string);
begin
  if Going then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWhileWorking, ['IP Restriction']));    {do not localize??}
  end;
  FIPRestriction := Value;
end;

procedure TIdHL7.SetPort(const Value: word);
begin
  if Going then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWhileWorking, ['Port']));          {do not localize??}
  end;
  FPort := Value;
end;

procedure TIdHL7.SetReconnectDelay(const Value: cardinal);
begin
  if Going then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWhileWorking, ['Reconnect Delay']));   {do not localize??}
  end;
  FReconnectDelay := Value;
end;

procedure TIdHL7.SetTimeOut(const Value: cardinal);
begin
  if Going then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWhileWorking, ['Time Out']));          {do not localize??}
  end;
  FTimeOut := Value;
end;

procedure TIdHL7.SetCommunicationMode(const Value: THL7CommunicationMode);
begin
  if Going then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWhileWorking, ['Communication Mode'])); {do not localize??}
  end;
  FCommunicationMode := Value;
end;

procedure TIdHL7.SetIsListener(const Value: boolean);
begin
  if Going then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWhileWorking, ['IsListener']));         {do not localize??}
  end;
  FIsListener := Value;
end;

function TIdHL7.GetStatus: TIdHL7Status;
begin
  FLock.Enter;
  try
    result := FStatus;
  finally
    FLock.Leave;
  end;
end;

function TIdHL7.Connected: boolean;
begin
  FLock.Enter;
  try
    result := FStatus = IsConnected;
  finally
    FLock.Leave;
  end;
end;

function TIdHL7.GetStatusDesc: string;
begin
  FLock.Enter;
  try
    result := FStatusDesc;
  finally
    FLock.Leave;
  end;
end;

procedure TIdHL7.InternalSetStatus(const AStatus: TIdHL7Status; ADesc: string);
begin
  FLock.Enter;
  try
    FStatus := AStatus;
    FStatusDesc := ADesc;
  finally
    FLock.Leave;
  end;
end;

{==========================================================
  Application Control
 ==========================================================}

procedure TIdHL7.Start;
var LStatus : TIdHL7Status;
begin
  LStatus := GetStatus;
  if LStatus = IsUnusable then
  begin
    raise EHL7CommunicationError.Create(Name, RSHL7NotFailedToStop);
  end;
  if LStatus <> IsStopped then
  begin
    raise EHL7CommunicationError.Create(Name, RSHL7AlreadyStarted);
  end;
  if FCommunicationMode = cmUnknown then
  begin
    raise EHL7CommunicationError.Create(Name, RSHL7ModeNotSet);
  end;
  if FCommunicationMode = cmAsynchronous then
  begin
    if not assigned(FOnMessageArrive) then
    begin
      raise EHL7CommunicationError.Create(Name, RSHL7NoAsynEvent);
    end;
  end;
  if (FCommunicationMode = cmSynchronous) and IsListener then
  begin
    if not assigned(FOnReceiveMessage) then
    begin
      raise EHL7CommunicationError.Create(Name, RSHL7NoSynEvent);
    end;
  end;
  FIsServer := (FAddress = '');
  if FIsServer  then
  begin
    StartServer
  end
  else
  begin
    StartClient;
  end;
  FPreStopped := false;
end;

procedure TIdHL7.PreStop;
  procedure JoltList(l:TList);
  var i:integer;
  begin
    for i := 0 to l.count -1 do
    begin
      TQueuedMessage(l[i]).FEvent.SetEvent;
    end;
  end;
begin
  if FCommunicationMode = cmSingleThread then
  begin
    FLock.Enter;
    try
      JoltList(FMsgQueue);
      JoltList(FHndMsgQueue);
    finally
      FLock.Leave;
    end;
  end;
  FPreStopped := true;
end;

procedure TIdHL7.Stop;
begin
  if not Going then
  begin
    raise EHL7CommunicationError.Create(Name, RSHL7AlreadyStopped);
  end;

  if not FPreStopped then
  begin
    PreStop;
    sleep(10); // give other threads a chance to clean up
  end;

  if FIsServer then
  begin
    StopServer
  end
  else
  begin
    StopClient;
  end;
end;


{==========================================================
  Server Connection Maintainance
 ==========================================================}

function TIdHL7.Going: boolean;
var LStatus:TIdHL7Status;
begin
  LStatus := GetStatus;
  result := (LStatus <> IsStopped) and (LStatus <> IsUnusable);
end;

{ TIdHL7PeerThread }

constructor TIdHL7PeerThread.Create(ACreateSuspended: boolean);
begin
  inherited;
  FBuffer := '';
end;

destructor TIdHL7PeerThread.Destroy;
begin
  inherited;
end;

procedure TIdHL7.CheckServerParameters;
begin
  if (FCommunicationMode = cmAsynchronous) or not FIsListener then
  begin
    FConnectionLimit := 1;
  end;

  if (FPort < 1) then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7InvalidPort, [FPort]));
  end;
end;

procedure TIdHL7.StartServer;
begin
  CheckServerParameters;
  FServer := TIdTCPServer.Create(nil);
  try
    FServer.DefaultPort := FPort;
    FServer.ThreadClass := TIdHL7PeerThread;
    Fserver.OnConnect := ServerConnect;
    FServer.OnExecute := ServerExecute;
    FServer.OnDisconnect := ServerDisconnect;
    FServer.Active := true;
    InternalSetStatus(IsNotConnected, RSHL7StatusNotConnected);
  except
    on e:exception do
    begin
      InternalSetStatus(IsStopped, Format(RSHL7StatusFailedToStart, [e.message]));
      FreeAndNil(FServer);
      raise;
    end;
  end;
end;

procedure TIdHL7.StopServer;
begin
  try
    FServer.Active := false;
    FreeAndNil(FServer);
    InternalSetStatus(IsStopped, RSHL7StatusStopped);
  except
    on e: Exception do
    begin
      InternalSetStatus(IsUnusable, Format(RSHL7StatusFailedToStop, [e.message]));
      FServer := nil;
      raise
    end;
  end;
end;

procedure TIdHL7.ServerConnect(AThread: TIdPeerThread);
var LNotify : boolean;
    LValid  : boolean;
begin
  FLock.Enter;
  try
    LNotify := FConnCount = 0;
    inc(FConnCount);
    // it would be better to stop getting here in the case of an invalid connection
    // cause here we drop it - nasty for the client. To be investigated later
    // can't use a ThreadPool - that interferes with server shutdown
    LValid  := FConnCount <= FConnectionLimit;
    If (FConnCount = 1) and (FCommunicationMode <> cmAsynchronous) and not IsListener then
    begin
      FServerConn := AThread.Connection;
    end;
    if LNotify then
    begin
      InternalSetStatus(IsConnected, RSHL7StatusConnected);
    end;
  finally
    FLock.Leave;
  end;
  if LValid then
  begin
    if LNotify and assigned(FOnConnect) then
    begin
      FOnConnect(self);
    end;
  end
  else
  begin
    // Thread exceeds connection limit
    AThread.Connection.Disconnect;
  end;
end;

procedure TIdHL7.ServerDisconnect(AThread: TIdPeerThread);
var LNotify : boolean;
begin
  FLock.Enter;
  try
    dec(FConnCount);
    LNotify := FConnCount = 0;
    if AThread.Connection = FServerConn then
    begin
      FServerConn := nil;
    end;
    if LNotify then
    begin
      InternalSetStatus(IsNotConnected, RSHL7StatusNotConnected);
    end;
  finally
    FLock.Leave;
  end;
  if LNotify and assigned(FOnDisconnect) then
  begin
    FOnDisconnect(self);
  end;
end;

procedure TIdHL7.ServerExecute(AThread: TIdPeerThread);
var LThread: TIdHL7PeerThread;
begin
  LThread := AThread as TIdHL7PeerThread;
  try
    // 1. prompt the network for content.
    LThread.Connection.ReadFromStack(false, -1, false);
  except
    try
      // well, there was some network error. We aren't sure what it
      // was, and it doesn't matter for this layer. we're just going
      // to make sure that we start again.
      // to review: what happens to the error messages?
      LThread.Connection.DisconnectSocket;
    except
      end;
    exit;
    end;
  if LThread.Connection.InputBuffer.Size > 0 then
    begin
    LThread.FBuffer := LThread.FBuffer + LThread.Connection.CurrentReadBuffer;
    HandleIncoming(LThread.FBuffer, LThread.Connection);
    end;
end;

procedure TIdHL7.DropServerConnection;
begin
  FLock.Enter;
  try
    if assigned(FServerConn) then
      FServerConn.Disconnect;
  finally
    FLock.Leave;
    end;
end;


{==========================================================
  Client Connection Maintainance
 ==========================================================}

procedure TIdHL7.CheckClientParameters;
begin
  if (FPort < 1) then
  begin
    raise EHL7CommunicationError.Create(Name, Format(RSHL7InvalidPort, [FPort]));
  end;
end;

procedure TIdHL7.StartClient;
begin
  CheckClientParameters;
  FClientThread := TIdHL7ClientThread.Create(self);
  InternalSetStatus(isConnecting, RSHL7StatusConnecting);
end;

procedure TIdHL7.StopClient;
var LFinished : boolean;
    LWaitStop : cardinal;
begin
  FLock.Enter;
  try
    FClientThread.Terminate;
    { this line is a problem - if the client interface thread is waiting
      reading from the network, then it will have an access violation - though
      this is handled internally, it is "unclean"}
    FClientThread.FClient.DisconnectSocket;
    FClientThread.FCloseEvent.SetEvent;
  finally
    FLock.Leave;
  end;
  LWaitStop := GetTickCount + 5000;
  repeat
    LFinished := (GetStatus = IsStopped);
    if not LFinished then
    begin
      sleep(10);
    end;
  until LFinished or (GetTickCount > LWaitStop);
  if GetStatus <> IsStopped then
  begin
    InternalSetStatus(IsUnusable, Format(RSHL7StatusFailedToStop, [RSHL7ClientThreadNotStopped]));
  end;
end;


procedure TIdHL7.DropClientConnection;
begin
  FLock.Enter;
  try
    if assigned(FClientThread) and assigned(FClientThread.FClient) then
      FClientThread.FClient.DisconnectSocket;
  finally
    FLock.Leave;
  end;
end;

{ TIdHL7ClientThread }

constructor TIdHL7ClientThread.Create(aOwner: TIdHL7);
begin
  FOwner := aOwner;
  FCloseEvent := TIdLocalEvent.Create(true, false);
  FreeOnTerminate := true;
  inherited Create(false);
end;

destructor TIdHL7ClientThread.Destroy;
begin
  FreeAndNil(FCloseEvent);
  try
    FOwner.FLock.Enter;
    try
      FOwner.FClientThread := nil;
      FOwner.InternalSetStatus(isStopped, RSHL7StatusStopped);
    finally
      FOwner.FLock.Leave;
    end;
  except
    // it's really vaguely possible that the owner
    // may be dead before we are. If that is the case, we blow up here.
    // who cares.
  end;
  inherited;
end;

procedure TIdHL7ClientThread.PollStack;
Var LBuffer : string;
begin
  LBuffer := '';
  repeat
    // we don't send here - we just poll the stack for content
    // if the application wants to terminate us at this point,
    // then it will disconnect the socket and we will get thrown
    // out
    // we really don't care at all whether the disconnect was clean or ugly

    // but we do need to suppress exceptions that come from
    // indy otherwise the client thread will terminate

    try
      // 1. prompt the network for content.
      FClient.ReadFromStack(false, -1, false);
    except
      try
        // well, there was some network error. We aren't sure what it
        // was, and it doesn't matter for this layer. we're just going
        // to make sure that we start again.
        // to review: what happens to the error messages?
        FClient.DisconnectSocket;
      except
        end;
      exit;
      end;
    if FClient.InputBuffer.Size > 0 then
      begin
      LBuffer := LBuffer + FClient.CurrentReadBuffer;
      FOwner.HandleIncoming(LBuffer, FClient);
      end;
  until Terminated or not FClient.Connected;
end;

procedure TIdHL7ClientThread.Execute;
var LRecTime : TDateTime;
begin
  try
    FClient := TIdTCPClient.Create(nil);
    try
      FClient.Host := FOwner.FAddress;
      FClient.Port := FOwner.FPort;
      repeat
        // try to connect. Try indefinitely but wait Owner.FReconnectDelay
        // between attempts. Problems: how long does Connect take?
        repeat
          FOwner.InternalSetStatus(IsConnecting, rsHL7StatusConnecting);
          try
            FClient.Connect;
          except
            on e: Exception do
            begin
              LRecTime := Now + ( (FOwner.FReconnectDelay / 1000) * {second length} (1 / ( 24 * 60 * 60)) );
              FOwner.InternalSetStatus(IsWaitReconnect, Format(rsHL7StatusReConnect, [FormatDateTime('hh:nn:ss', LRecTime), e.message])); {do not localize??}
            end;
          end;
          if not Terminated and not FClient.Connected then
          begin
            {$IFDEF SYNCOBJS_WORKAROUND}
            FCloseEvent.WaitFor;
            {$ELSE}
            FCloseEvent.WaitFor(FOwner.FReconnectDelay);
            {$ENDIF}
          end;
        until Terminated or FClient.Connected;
        if Terminated then
        begin
          exit;
        end;

        FOwner.FLock.Enter;
        try
          FOwner.FClient := FClient;
          FOwner.InternalSetStatus(IsConnected, rsHL7StatusConnected);
        finally
          FOwner.FLock.Leave;
        end;
        if assigned(FOwner.FOnConnect) then
        begin
          FOwner.FOnConnect(FOwner);
        end;
        try
          PollStack;
        finally
          FOwner.FLock.Enter;
          try
            FOwner.FClient := nil;
            FOwner.InternalSetStatus(IsNotConnected, RSHL7StatusNotConnected);
          finally
            FOwner.FLock.Leave;
          end;
          if assigned(FOwner.FOnDisconnect) then
          begin
            FOwner.FOnDisconnect(FOwner);
          end;
        end;
        if not Terminated then
        begin
          // we got disconnected. ReconnectDelay applies.
          {$IFDEF SYNCOBJS_WORKAROUND}
          FCloseEvent.WaitFor;
          {$ELSE}
          FCloseEvent.WaitFor(FOwner.FReconnectDelay);
          {$ENDIF}
        end;
      until terminated;
    finally
      FreeAndNil(FClient);
    end;
  except
    on e: Exception do
      // presumably some comms or indy related exception
      // there's not really anyplace good to put this????
  end;
end;

{==========================================================
  Internal process management
 ==========================================================}

procedure TIdHL7.HandleIncoming(Var VBuffer : String; AConnection : TIdTCPConnection);
var LStart, LEnd : integer;
    LMsg, LReply : string;
begin
  try
    // process any messages in the buffer (may get more than one per packet)
    repeat
      { use of Pos instead of Indypos is deliberate }
      LStart := pos(MSG_START, VBuffer);
      LEnd   := pos(MSG_END, VBuffer);
      if (LStart > 0) and (LEnd > 0) then
      begin
        LMsg := copy(VBuffer, LStart+length(MSG_START), LEnd - (LStart + length(MSG_START)));
        delete(VBuffer, 1, (LEnd-1)+ length(MSG_END));
        if HandleMessage(LMsg, LReply) then
        begin
          if LReply <> '' then
          begin
            AConnection.Write(MSG_START+LReply+MSG_END);
          end;
        end
        else
        begin
          AConnection.DisconnectSocket;
        end;
      end;
    until (LEnd = 0);
    if length(VBuffer) > BUFFER_SIZE_LIMIT then
    begin
      AConnection.DisconnectSocket;
    end;
  except
    // well, we need to suppress the exception, and force a reconnection
    // we don't know why an exception has been allowed to propagate back
    // to us, it shouldn't be allowed. so what we're going to do, is drop
    // the connection so that we force all the network layers on both
    // ends to reconnect.
    // this is a waste of time of the error came from the application but
    // this is not supposed to happen
    try
      AConnection.DisconnectSocket;
    except
      // nothing - suppress
      end;
    end;
end;

function  TIdHL7.HandleMessage(const AMsg: string; var VReply: string): boolean;
var LQueMsg  : TQueuedMessage;
    LIndex   : integer;
begin
  VReply := '';
  try
    case FCommunicationMode of
      cmUnknown      :
         begin
         raise EHL7CommunicationError.Create(Name, RSHL7ImpossibleMessage);
         end;
      cmAsynchronous :
         begin
         FOnMessageArrive(self, Amsg);
         end;
      cmSynchronous, cmSingleThread :
         begin
         if IsListener then
         begin
           if FCommunicationMode = cmSynchronous then
           begin
             FOnReceiveMessage(self, AMsg, VReply)
           end
           else
           begin
             LQueMsg := TQueuedMessage.Create(AMsg, FReceiveTimeout);
             try
               FLock.Enter;
               try
                 FMsgQueue.Add(LQueMsg);
               finally
                 FLock.Leave;
                 end;
               LQueMsg.wait;
               // no locking. There is potential problems here. To be reviewed
               VReply := LQueMsg.FReply;
             finally
               FLock.Enter;
               try
                 LIndex := FMsgQueue.IndexOf(LQueMsg);
                 if LIndex > -1 then
                   FMsgQueue.Delete(LIndex);
               finally
                 FLock.Leave;
                 end;
               FreeAndNil(LQueMsg);
               end;
           end
         end
         else
         begin
           FLock.Enter;
           try
             if FWaitingForAnswer then
             begin
               FWaitingForAnswer := false;
               FMsgReply := AMsg;
               FReplyResponse := srOK;
               if FCommunicationMode = cmSynchronous then
               begin
                 FWaitEvent.SetEvent;
               end;
             end
             else
             begin
               // we could have got here by timing out, but this is quite unlikely,
               // since the connection will be dropped in that case. We will report
               // this as a spurious message
               raise EHL7CommunicationError.Create(Name, RSHL7UnexpectedMessage);
             end;
           finally
             FLock.Leave;
             end;
         end
         end;
      else
        begin
        raise EHL7CommunicationError.Create(Name, RSHL7UnknownMode);
        end;
     end;
  except
    on e: Exception do
      if Assigned(FOnReceiveError) then
      begin
        FOnReceiveError(self, AMsg, e, VReply, result)
      end
      else
      begin
        result := false;
      end;
  end;
end;

{==========================================================
  Sending
 ==========================================================}

// this procedure is not technically thread safe.
// if the connection is disappearing when we are attempting
// to write, we can get transient access violations. Several
// strategies are available to prevent this but they significantly
// increase the scope of the locks, which costs more than it gains

function TIdHL7.AsynchronousSend(AMsg: string): TSendResponse;
begin
  result := srNone; // just to suppress the compiler warning
  FLock.Enter;
  try
    if not Going then
    begin
      raise EHL7CommunicationError.Create(Name, Format(RSHL7NotWorking, [RSHL7SendMessage]))
    end
    else if GetStatus <> isConnected then
    begin
      result := srNoConnection
    end
    else
    begin
      if FIsServer then
      begin
        if Assigned(FServerConn) then
        begin
          FServerConn.Write(MSG_START+AMsg+MSG_END);
          result := srSent
        end
        else
        begin
          raise EHL7CommunicationError.Create(Name, RSHL7NoConnectionFound);
        end
      end
      else
      begin
        FClient.Write(MSG_START+AMsg+MSG_END);
        result := srSent
      end;
    end;
  finally
    FLock.Leave;
  end
end;

function TIdHL7.SynchronousSend(AMsg: string; var VReply: string): TSendResponse;
begin
  result := srError;
  FLock.Enter;
  try
    FWaitingForAnswer := true;
    FReplyResponse := srTimeout;
    FMsgReply := '';
  finally
    FLock.Leave;
  end;
  try
    result  := AsynchronousSend(AMsg);
    if result = srSent then
    begin
      {$IFDEF SYNCOBJS_WORKAROUND}
      FWaitEvent.WaitFor;
      {$ELSE}
      FWaitEvent.WaitFor(FTimeOut);
      {$ENDIF}
    end;
  finally
    FLock.Enter;
    try
      FWaitingForAnswer := false;
      if result = srSent then
      begin
        result := FReplyResponse;
      end;
      if result = srTimeout then
        begin
        if FIsServer then
          DropServerConnection
        else
          DropClientConnection;
        end;
      VReply := FMsgReply;
    finally
      FLock.Leave;
    end;
  end;
end;

procedure TIdHL7.SendMessage(AMsg: string);
begin
  if FWaitingForAnswer then
    raise EHL7CommunicationError.Create(Name, RSHL7WaitForAnswer);

  FLock.Enter;
  try
    FWaitingForAnswer := true;
    FMsgReply := '';
    FReplyResponse := AsynchronousSend(AMsg);
  finally
    FLock.Leave;
  end;
end;

function TIdHL7.GetReply(var VReply: string): TSendResponse;
begin
  FLock.Enter;
  try
    if FWaitingForAnswer then
    begin
      result := srNone
    end
    else
    begin
      result := FReplyResponse;
      if result = srSent then
      begin
        result := srTimeOut;
      end;
      VReply  := FMsgReply;
      FWaitingForAnswer := false;
      FReplyResponse := srError;
    end;
  finally
    FLock.Leave;
  end;
end;

function TIdHL7.GetMessage(var VMsg: string): pointer;
begin
  FLock.Enter;
  try
    if FMsgQueue.count = 0 then
    begin
      result := nil
    end
    else
    begin
      result := FMsgQueue[0];
      VMsg := TQueuedMessage(result).FMsg;
      FMsgQueue.Delete(0);
      FHndMsgQueue.Add(result);
    end;
  finally
    FLock.Leave;
  end;
end;

procedure TIdHL7.SendReply(AMsgHnd: pointer; AReply: string);
var qm : TQueuedMessage;
begin
  FLock.Enter;
  try
    qm := TObject(AMsgHnd) as TQueuedMessage;
    qm.FReply := AReply;
    FHndMsgQueue.Delete(FHndMsgQueue.IndexOf(AMsgHnd));
  finally
    FLock.Leave;
  end;
  qm.FEvent.SetEvent;
end;

end.
